from dune.perftool.ufl.modified_terminals import Restriction

from dune.perftool.options import get_option

from dune.perftool.pdelab.argument import name_coefficientcontainer

from dune.perftool.generation import (class_member,
                                      constructor_block,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      include_file,
                                      initializer_list,
                                      silenced_warning,
                                      temporary_variable,
                                      valuearg
                                      )
from dune.perftool.loopy.buffer import get_buffer_temporary
from dune.perftool.pdelab.localoperator import (name_domain_field,
                                                lop_template_range_field,
                                                )
from dune.perftool.pdelab.quadrature import quadrature_order
from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pytools import Record, product

import numpy


class AMatrix(Record):
    def __init__(self, a_matrix, rows, cols):
        Record.__init__(self,
                        a_matrix=a_matrix,
                        rows=rows,
                        cols=cols,
                        )


class ColMajorAccess(FunctionIdentifier):
    def __init__(self, amatrix):
        self.amatrix = amatrix

    def __getinitargs__(self):
        return (self.amatrix,)

    @property
    def name(self):
        return '{}.colmajoraccess'.format(self.amatrix)


@function_mangler
def colmajoraccess_mangler(target, func, dtypes):
    if isinstance(func, ColMajorAccess):
        return CallMangleInfo(func.name, (NumpyType(numpy.float64),), (NumpyType(numpy.int32), NumpyType(numpy.int32)))


@class_member("operator")
def define_alignment(name):
    alignment = get_option("sumfact_alignment")
    return "enum {{ {} = {} }};".format(name, str(alignment))


def name_alignment():
    name = "alignment"
    define_alignment(name)
    return name


def quadrature_points_per_direction():
    # TODO use quadrature order from dune.perftool.pdelab.quadrature
    # as soon as we have a generic implementation

    # Quadrature order
    q = quadrature_order()

    # Quadrature points in per direction
    nb_qp = q // 2 + 1

    return nb_qp


@class_member("operator")
def define_number_of_quadrature_points_per_direction(name):
    number_qp = quadrature_points_per_direction()
    return "enum {{ {} = {} }};".format(name, str(number_qp))


def name_number_of_quadrature_points_per_direction():
    name = "m"
    define_number_of_quadrature_points_per_direction(name)
    return name


def polynomial_degree():
    form = get_global_context_value("formdata").preprocessed_form
    return form.coefficients()[0].ufl_element()._degree


def basis_functions_per_direction():
    return polynomial_degree() + 1


@class_member("operator")
def define_number_of_basis_functions_per_direction(name):
    number_basis = basis_functions_per_direction()
    return "enum {{ {} = {} }};".format(name, str(number_basis))


def name_number_of_basis_functions_per_direction():
    name = "n"
    define_number_of_basis_functions_per_direction(name)
    return name


@class_member("operator")
def define_oned_quadrature_weights(name):
    range_field = lop_template_range_field()
    number_qp = name_number_of_quadrature_points_per_direction()
    return "{} {}[{}];".format(range_field, name, number_qp)


def name_oned_quadrature_weights():
    name = "qw"
    globalarg(name, shape=(quadrature_points_per_direction(),), dtype=NumpyType(numpy.float64))
    define_oned_quadrature_weights(name)
    return name


@class_member("operator")
def define_oned_quadrature_points(name):
    range_field = lop_template_range_field()
    number_qp = name_number_of_quadrature_points_per_direction()
    return "{} {}[{}];".format(range_field, name, number_qp)


def name_oned_quadrature_points():
    name = "qp"
    globalarg(name, shape=(quadrature_points_per_direction(),), dtype=NumpyType(numpy.float64))
    define_oned_quadrature_points(name)
    return name


@class_member("operator")
def typedef_polynomials(name):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    degree = polynomial_degree()

    include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="operatorfile")

    # TODO: make switching polynomials possible
    # return "using {} = Dune::QkStuff::GaussLobattoLagrangePolynomials<{}, {}, {}>;".format(name,
    #                                                                                        domain_field,
    #                                                                                        range_field,
    #                                                                                        degree)
    return "using {} = Dune::QkStuff::EquidistantLagrangePolynomials<{}, {}, {}>;".format(name,
                                                                                          domain_field,
                                                                                          range_field,
                                                                                          degree)


def type_polynomials():
    name = "Polynomials1D"
    typedef_polynomials(name)
    return name


@class_member("operator")
def define_polynomials(name):
    polynomials_type = type_polynomials()
    return "{} {};".format(polynomials_type, name)


def name_polynomials():
    name = "poly"
    define_polynomials(name)
    return name


@constructor_block("operator")
def sort_quadrature_points_weights():
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    number_qp = name_number_of_quadrature_points_per_direction()
    qp = name_oned_quadrature_points()
    qw = name_oned_quadrature_weights()
    include_file("dune/perftool/sumfact/onedquadrature.hh", filetag="operatorfile")
    return "onedQuadraturePointsWeights<{}, {}, {}>({}, {});".format(range_field, domain_field, number_qp, qp, qw)


@constructor_block("operator")
def construct_theta(name, transpose, derivative):
    # Make sure that the quadrature points are sorted
    sort_quadrature_points_weights()

    if transpose:
        shape = (name_number_of_basis_functions_per_direction(), name_number_of_quadrature_points_per_direction())
    else:
        shape = (name_number_of_quadrature_points_per_direction(), name_number_of_basis_functions_per_direction())
    polynomials = name_polynomials()
    qp = name_oned_quadrature_points()

    # access = "j,i" if transpose else "i,j"
    basispol = "dp" if derivative else "p"
    polynomial_access = "i,{}[j]".format(qp) if transpose else "j,{}[i]".format(qp)

    return ["for (std::size_t i=0; i<{}; i++){{".format(shape[0]),
            "  for (std::size_t j=0; j<{}; j++){{".format(shape[1]),
            "    {}.colmajoraccess(i,j) = {}.{}({});".format(name, polynomials, basispol, polynomial_access),
            "  }",
            "}"]


@class_member("operator")
def typedef_theta(name):
    include_file("dune/perftool/sumfact/alignedmatvec.hh", filetag="operatorfile")
    alignment = name_alignment()
    range_field = lop_template_range_field()
    return "using {} = AlignedMatrix<{}, {}>;".format(name, range_field, alignment)


def type_theta():
    name = "AMatrix"
    typedef_theta(name)
    return name


@class_member("operator")
def define_theta(name, shape, transpose, derivative):
    theta_type = type_theta()
    initializer_list(name, [str(axis) for axis in shape], classtag="operator")
    construct_theta(name, transpose, derivative)
    return "{} {};".format(theta_type, name)


def name_theta():
    name = "Theta"
    shape = (quadrature_points_per_direction(), basis_functions_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, False, False)
    return name


def name_theta_transposed():
    name = "ThetaT"
    shape = (basis_functions_per_direction(), quadrature_points_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, True, False)
    return name


def name_dtheta():
    name = "dTheta"
    shape = (quadrature_points_per_direction(), basis_functions_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, False, True)
    return name


def name_dtheta_transposed():
    name = "dThetaT"
    shape = (basis_functions_per_direction(), quadrature_points_per_direction())
    globalarg(name, shape=shape, dtype=numpy.float64, dim_tags="f,f")
    define_theta(name, shape, True, True)
    return name
